import unittest
import filecmp
import tempfile
import os
import sys

sys.path = [os.path.join(os.path.dirname(
    os.path.realpath(__file__)), '..', )]+sys.path

from enrichm.writer import Writer

class Tests(unittest.TestCase):

    def test_write(self):
        output_file = tempfile.mktemp()
        output_lines = [["this", "is", "a", "header"], ["Content", 1, 2, 3]]
        expected_list = ["this\tis\ta\theader\n", "Content\t1\t2\t3\n"]

        Writer.write(output_lines, output_file)

        with open(output_file) as output_file_io:
            for idx, line in enumerate(output_file_io):
                self.assertEqual(line, expected_list[idx])
                
            self.assertEqual(idx+1, len(expected_list))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import logging
import os
from itertools import chain
from collections import Counter
from enrichm.databases import Databases
# Local
###############################################################################
class Writer:
    '''
    A collection of functions to write files in various formats.
    '''

    @staticmethod
    def write(output_lines_list, output_path):
        '''

        Parameters
        ----------
        output_lines_list   - List. A list of lists. Each sublist is a line,
                              where each entry is a column entry
        output_path         - String. Path to write output lines to.
        Output
        ------
        '''
        logging.info("Writing results to file: %s" % output_path)

        with open(output_path, 'w') as out_io:

            for output_line_list in output_lines_list:
                output_line_string = '\t'.join([str(column_entry) for column_entry in output_line_list]) + '\n'
                out_io.write(output_line_string)

            out_io.flush()
            out_io.close()

    @staticmethod
    def write_gff(genome, output_file):
        '''
        Writes a gff file for a Genome object

        Parameters
        ----------
        genome 			- Genome object
        output_file 	- string. file name to output results to.
        '''
        with open(output_file, 'w') as out_io:
            for sequence in genome.ordered_sequences():
                contig_id = '_'.join(sequence.seqname.split('_')[:-1])

                features = ['seq_id=%s' % sequence.seqname,
                                    'prodigal_id=%s' % sequence.prod_id,
                                    'partial=%s' % sequence.partial,
                                    'start_type=%s' % sequence.starttype,
                                    'rbs_motif=%s' % sequence.rbs_motif,
                                    'rbs_spacer=%s' % sequence.rbs_spacer,
                                    'gc=%s' % sequence.gc]

                if len(sequence.all_annotations()) > 0:
                    features.append('annotations=%s' % ','.join(sequence.all_annotations()))
                else:
                    features.append('annotations=hypothetical_protein')

                line = [contig_id,
                                    'prodigal',
                                    'CDS',
                                    sequence.startpos,
                                    sequence.finishpos,
                                    '.',
                                    ('-' if sequence.direction == '-1'
                                     else '+'),
                                    '0',
                                    ';'.join(features)]

                out_io.write('\t'.join(line) + '\n')

class MatrixGenerator:

    KO      = 'KO_IDS.txt'
    EC      = 'EC_IDS.txt'
    PFAM    = 'PFAM_IDS.txt'
    TIGRFAM = 'TIGRFAM_IDS.txt'
    CAZY = 'CAZY_IDS.txt'
    HYPOTHETICAL = 'HYPOTHETICAL'
    ORTHOLOG = 'ORTHOLOG'

    def __init__(self, annotation_type, clusters = None):
        '''
        Interpret which annotation type to write a matrix for.

        Parameters
        ----------
        annotation_type - String.
        '''
        self.annotation_type = annotation_type
        self.databases = Databases()
        if self.annotation_type == self.KO:
            self.annotation_list = [x.strip() for x in open(os.path.join(self.databases.IDS_DIR, self.KO))]

        elif self.annotation_type == self.EC:
            self.annotation_list = [x.strip() for x in open(os.path.join(self.databases.IDS_DIR, self.EC))]

        elif self.annotation_type == self.PFAM:
            self.annotation_list = [x.strip() for x in open(os.path.join(self.databases.IDS_DIR, self.PFAM))]

        elif self.annotation_type == self.TIGRFAM:
            self.annotation_list = [x.strip() for x in open(os.path.join(self.databases.IDS_DIR, self.TIGRFAM))]

        elif self.annotation_type == self.CAZY:
            self.annotation_list = [x.strip() for x in open(os.path.join(self.databases.IDS_DIR, self.CAZY))]

        elif self.annotation_type == self.HYPOTHETICAL:
            self.annotation_list = clusters

        elif self.annotation_type == self.ORTHOLOG:
            self.annotation_list = clusters

        else:
            raise Exception("Annotation type not found: %s" % (self.annotation_type))

    def write_matrix(self, genomes_list, count_domains, output_path):
        '''
        Writes a frequency matrix with of each annotation (rows) per sample (columns)

        Parameters
        ----------
        genomes_list        - list. List of Genome objects
        output_path         - string. Path to file to which the results are written.
        '''

        logging.info("    - Writing results to file: %s" % output_path)

        with open(output_path, 'w') as out_io:
            colnames = ['ID'] + [genome.name for genome in genomes_list]
            out_io.write('\t'.join(colnames) + '\n')

            if count_domains:
                genome_annotations = {genome.name:Counter(chain(*[sequence.all_annotations() for sequence in genome.sequences.values()]))
                                      for genome in genomes_list}
            else:
                genome_annotations = {genome.name:Counter(chain(*[set(sequence.all_annotations()) for sequence in genome.sequences.values()]))
                                      for genome in genomes_list}

            for annotation in self.annotation_list:
                output_line = [annotation]

                for genome in genomes_list:

                    if annotation in genome_annotations[genome.name]:
                        output_line.append(str(genome_annotations[genome.name][annotation]))

                    else:
                        output_line.append('0')

                out_io.write( '\t'.join(output_line) + '\n' )



#!/usr/bin/env python3
# Imports
import logging
import re
from itertools import product
# Local
from enrichm.databases import Databases
from enrichm.parser import Parser
###############################################################################

class NetworkBuilder:

    MODULE_PREFIX   = 'M'
    MAP_PREFIX      = 'map'
    PATHWAY_PREFIX  = 'rn'
    REACTION_PREFIX = 'R'
    ZERO            = '0.0'

    def __init__(self, metadata, abundances_metagenome, abundances_transcriptome, abundances_metabolome, fisher_results):
        self.abundances_metagenome = abundances_metagenome
        self.abundances_transcriptome = abundances_transcriptome
        self.abundances_metabolome = abundances_metabolome
        self.fisher_results = fisher_results
        self.metadata_keys = list(metadata.keys())

        databases = Databases()

        self.reaction_to_module = databases.r2m()
        self.module_to_reaction = databases.m2r()
        self.module_descriptions = databases.m()
        self.reaction_to_pathway = databases.r2p()
        self.pathway_to_reaction = databases.p2r()
        self.pathway_descriptions = databases.p()
        self.compound_desc_dict = databases.compound_desc_dict()
        self.compound_descriptions = databases.c()
        self.reaction_descriptions = databases.r()
        self.reactions_to_compounds = databases.r2c()
        self.reactions_to_kos = databases.r2k()

        self.matrix_header = ["compound", "reaction", 'type']
        self.transcriptome_header = [key + '_reaction_transcriptome' for key in self.metadata_keys]
        self.compound_header = [key + '_compound' for key in self.metadata_keys]
        self.metadata_header = ['node', 'description', 'type', 'module', 'module_descr', 'pathway', 'pathway_descr', 'node_type']
        self.query_header =  ['query', 'step']
        self.step_header = ['step']
        self.to_omit = set(["C00828",  # Menaquinone
                            "C00534",  # Pyridoxamine
                            "C00006",  # NADP+
                            "C00003",  # NAD+
                            "C00002",  # ATP
                            "C00314",  # Pyridoxine
                            "C00864",  # Pantothenate
                            "C00504",  # Folate
                            "C00032",  # Heme
                            "C05443",  # Vitamin D3
                            "C00253",  # Nicotinate
                            "C00250",  # Pyridoxal
                            "C11378",  # Ubiquinone-10
                            "C05777",  # Coenzyme F430
                            "C00072",  # Ascorbate
                            "C00378",  # Thiamine
                            "C00101",  # Tetrahydrofolate
                            "C00029",  # UDP-glucose
                            "C00068",  # Thiamin diphosphate
                            "C00061",  # FMN
                            "C00063",  # CTP
                            "C05776",  # Vitamin B12
                            "C00113",  # PQQ
                            "C18237",  # Molybdoenzyme molybdenum cofactor
                            "C00051",  # Glutathione
                            "C00010",  # CoA
                            "C00016",  # FAD
                            "C00018",  # Pyridoxal phosphate
                            "C00019",  # S-Adenosyl-L-methionine
                            "C00153",  # Nicotinamide
                            "C04628",  # Coenzyme B
                            "C00862",  # Methanofuran
                            "C15672",  # Heme O
                            "C15670",  # Heme A
                            "C02059",  # Phylloquinone
                            "C03576",  # Coenzyme M
                            "C05441",  # Vitamin D2
                            "C00272",  # Tetrahydrobiopterin
                            "C02477",  # alpha-Tocopherol
                            "C00473",  # Retinol
                            "C00120",  # Biotin
                            "C00725",  # Lipoate
                            "C00053",  # 3'-Phosphoadenylyl sulfate
                            "C00194",  # Cobamide coenzyme
                            "C00255",  # Riboflavin
                            'C00001',  # H2O
                            'C00008',  # ADP
                            'C00013',  # Diphosphate
                            'C00004',  # NADH
                            'C00005',  # NADPH
                            'C00080',  # H+
                            'C00009',  # Orthophosphate
                            'C00008',  # ADP
                            'C00004',  # NADH
                            'C00020',  # AMP
                            'C00007',  # Oxygen
                            'C00015']) # UDP

    def nested_dict_vals(self, input_dictionary):
        reaction_regex = '(R\d{5})$'

        for key, item in input_dictionary.items():

            if isinstance(item, dict):
                yield from self.nested_dict_vals(item)
            else:

                if type(reaction_regex) == str:

                    if re.match(reaction_regex, key):
                        yield key

                elif type(reaction_regex) == bytes:
                    key = key.decode()

                    if re.match(reaction_regex, key):
                        yield key

                else:
                    raise Exception("Invalid key in in nested dict!")

    def gather_module(self, key):

        if key in self.reaction_to_module:
            len_list =  [len(self.module_to_reaction[x]) for x in self.reaction_to_module[key]]
            module = self.reaction_to_module[key][len_list.index(max(len_list))]
            module_description = self.module_descriptions[module]
        else:
            module = 'NA'
            module_description='NA'

        return module, module_description

    def gather_pathway(self, key):

        if key in self.reaction_to_pathway:
            len_list =  [len(self.pathway_to_reaction[x]) for x in self.reaction_to_pathway[key]]
            pathway = self.reaction_to_pathway[key][len_list.index(max(len_list))]
            pathway_description = self.pathway_descriptions[pathway]
        else:
            pathway = 'NA'
            pathway_description='NA'

        return pathway, pathway_description

    def gather_compound_metadata(self, compound):

        if compound in self.compound_desc_dict:
            compound_type = ','.join(self.compound_desc_dict[compound]['A'])
        else:
            compound_type = 'NA'

        compound_metadata_list = [compound, self.compound_descriptions[compound], compound_type, 'NA', 'NA', 'NA', 'NA', 'compound']

        if self.abundances_metabolome:

            for sample in self.metadata_keys:

                if compound in self.abundances_metabolome[sample]:
                    compound_abundance = self.abundances_metabolome[sample][compound]
                else:
                    compound_abundance = 0

                compound_metadata_list.append(compound_abundance)

        return compound_metadata_list

    def gather_reaction_metadata(self, reaction):
        module, module_description = self.gather_module(reaction)
        pathway, pathway_description = self.gather_pathway(reaction)
        reaction_metadata_list = [reaction, self.reaction_descriptions[reaction], 'NA', module, module_description, pathway, pathway_description, 'reaction']

        if self.abundances_metabolome:
            reaction_metadata_list += ['-5' for sample in self.metadata_keys]

        return reaction_metadata_list

    def gather_enriched_term(self, reaction, fisher_results):
        enriched_term = list()

        for compared_group in list(fisher_results.keys()):

            if any(set(self.reactions_to_kos[reaction]).intersection(fisher_results[compared_group])):
                enriched_term.append(compared_group)

        if len(enriched_term) > 0:
            enriched_term = '_'.join(enriched_term)
        else:
            enriched_term = 'NA'

        return enriched_term

    def get_reaction_compounds(self, compounds_list):

        for compound in compounds_list:

            if compound not in self.to_omit:
                yield compound

    def gather_reaction_edge_data(self, compound, reaction):

        if self.fisher_results:
            enriched_term = self.gather_enriched_term(reaction, self.fisher_results)
        else:
            enriched_term = 'NA'

        reaction_line = [compound, reaction, enriched_term]

        for key in self.metadata_keys:

            for _, group_abundances in self.abundances_metagenome.items():
                if reaction in group_abundances[key]:
                    reaction_line.append(str(group_abundances[key][reaction]))
                else:
                    reaction_line.append(self.ZERO)

        if self.abundances_transcriptome:

            for key in self.metadata_keys:
                for _, group_abundances in self.abundances_transcriptome.items():
                    if reaction in group_abundances[key]:
                        reaction_line.append(str(group_abundances[key][reaction]))
                    else:
                        reaction_line.append(self.ZERO)

        if sum([float(x) for x in reaction_line[3:]])>0:
            return reaction_line

    def add_to_header(self, groups):
        output_list = list()
        groups = list(groups.keys())

        for genome_group, reference_group in product(self.metadata_keys, groups):
            output_list.append(f"{genome_group}_{reference_group}")
        
        return output_list

    def all_matrix(self, reference_dict):
        '''
        Parameters
        ----------
        '''
        seen_nodes = set()
        
        # Construct headers to network matrices
        seen_reactions = set(self.nested_dict_vals(self.abundances_metagenome))

        network_lines = [self.matrix_header]

        if self.abundances_metagenome:
            network_lines[0] += self.add_to_header(self.abundances_metagenome)
        if self.abundances_transcriptome:
            network_lines[0] += self.add_to_header(self.abundances_transcriptome)
        if self.abundances_metabolome:
            node_metadata_lines = [self.metadata_header + self.compound_header]
        else:
            node_metadata_lines = [self.metadata_header]
        
        for reaction, entry in reference_dict.items():

            for compound in entry:

                if reaction in seen_reactions:

                    #~#~#~#~#~#~#~#~ Fill in abundances_metagenome ~#~#~#~#~#~#~#~
                    reaction_line = self.gather_reaction_edge_data(compound, reaction)

                    if reaction_line:
                        network_lines.append(reaction_line)

                    #~#~#~#~#~#~#~#~#~ Fill in node metadata ~#~#~#~#~#~#~#~#~#
                    if compound not in seen_nodes:
                        compound_metadata_list = self.gather_compound_metadata(compound)

                        node_metadata_lines.append(compound_metadata_list)
                        seen_nodes.add(compound)

                    if reaction not in seen_nodes:
                        reaction_metadata_list = self.gather_reaction_metadata(reaction)

                        node_metadata_lines.append(reaction_metadata_list)
                        seen_nodes.add(reaction)

        return network_lines, node_metadata_lines

    def query_matrix(self, queries, depth):
        '''
        Parameters
        ----------
        '''
        steps = 0
        queries_list = Parser.parse_single_column_text_file(queries)
        seen_reactions = set(self.nested_dict_vals(self.abundances_metagenome))
        seen_nodes = set()
        level_queries = set()

        node_metadata_lines = [self.metadata_header + self.query_header]

        if self.abundances_metabolome:
            node_metadata_lines[0] += self.compound_header

        network_lines = [self.matrix_header]
        if self.abundances_metagenome:
            network_lines[0] += self.add_to_header(self.abundances_metagenome)
        if self.abundances_transcriptome:
            network_lines[0] += self.add_to_header(self.abundances_transcriptome)
        network_lines[0] += self.step_header

        while depth>0:

            if any(level_queries):
                queries_list = set(level_queries)
                level_queries = set()

            for reaction, entry in self.reactions_to_compounds.items():

                if reaction in seen_reactions:

                    if any(queries_list.intersection(entry)):

                        for compound in self.get_reaction_compounds(entry):

                            #~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#
                            #~#~#~#~#~#~#~#~ Fill in abundances ~#~#~#~#~#~#~#~
                            reaction_line = self.gather_reaction_edge_data(compound, reaction)

                            if reaction_line:
                                reaction_line.append(str(steps))
                                network_lines.append(reaction_line)

                            #~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#
                            #~#~#~#~#~#~#~ Fill in node metadata  ~#~#~#~#~#~#~
                            index = str((steps if compound in queries_list else steps+1))

                            if compound not in seen_nodes:

                                metadata_list = self.gather_compound_metadata(compound)

                                if compound in queries_list:
                                    is_query = 'True'
                                else:
                                    is_query = 'False'

                                node_metadata_lines.append(metadata_list + [is_query, index])
                                seen_nodes.add(compound)

                                if compound not in queries_list:
                                    level_queries.add(compound)

                            if reaction not in seen_nodes:
                                metadata_list = self.gather_reaction_metadata(reaction)

                                node_metadata_lines.append(metadata_list + ['False', index])
                                seen_nodes.add(reaction)

            steps+=1
            depth-=1

            logging.info("Step %i complete with %i queries to continue with" % (steps, len(level_queries)))

        return network_lines, node_metadata_lines

    def pathway_matrix(self, limit, filter):

        possible_reactions = set()

        if any(limit):

            for entry in limit:

                if(entry.startswith(self.MAP_PREFIX) or
                   entry.startswith(self.PATHWAY_PREFIX)):

                    for reaction in self.pathway_to_reaction[entry]:
                        possible_reactions.add(reaction)

                elif(entry.startswith(self.MODULE_PREFIX)):

                    for reaction in self.module_to_reaction[entry]:
                        possible_reactions.add(reaction)

                elif(entry.startswith(self.REACTION_PREFIX)):
                    possible_reactions.add(entry)

            possible_reactions = {reaction:self.reactions_to_compounds[reaction]
                                  for reaction in
                                  possible_reactions
                                  if reaction in self.reactions_to_compounds}
        else:
            possible_reactions = self.reactions_to_compounds

        for entry in filter:

            if entry in possible_reactions:

                del possible_reactions[entry]

        network_lines, node_metadata_lines = self.all_matrix(possible_reactions)
        return network_lines, node_metadata_lines




#!/usr/bin/env python
# Imports
import unittest
import os.path
import sys
import subprocess
import tempfile
import filecmp
from scipy import stats


path_to_script = os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','bin','enrichm')
path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data')
path_to_annotate = os.path.join(path_to_data, 'enrichm_annotate')

sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..')]+sys.path

from enrichm.enrichment import Test
from enrichm.databases import Databases

###############################################################################

class Tests(unittest.TestCase):
    
    genome_annotation_simple_example = {"genome_1": {"K00001":1,
                                                     "K00002":2,
                                                     "K00003":0},
                                        "genome_2": {"K00001":0,
                                                     "K00002":0,
                                                     "K00003":1},
                                        "genome_3": {"K00001":5,
                                                     "K00002":4,
                                                     "K00003":5}}
    genome_groups_simple_example = {"group_1": ["genome_1"],
                                    "group_2": ["genome_2", "genome_3"]}
    simple_test_object = Test(genome_annotation_simple_example,
                              genome_groups_simple_example,
                              "kegg",
                              0.1,
                              'fdr_bh',
                              1,
                              Databases())
    sample_to_annotation = {'sample_group_1': {'K00001': [16.0, 25.5, 30.1],
                                               'K00002': [14.0, 21.0, 24.2],
                                               'K00003': [15.5, 26.2, 31.1]},
                            'sample_group_2': {'K00001': [6.0, 6.5, 7.0],
                                               'K00002': [10.8, 12.4, 14.0],
                                               'K00003': [6.2, 5.4, 5.0]}}
    annotations = ["K00001", "K00002", "K00003"]

    def test_test_chooser(self):
        groups_1 = [[1, 2, 3, 4, 5], [1, 2, 3, 4, 5]]
        groups_2 = [[1], [1, 2, 3, 4, 5]]
                
        test_instance_1 = self.simple_test_object.test_chooser(groups_1)
        test_instance_2 = self.simple_test_object.test_chooser(groups_2)
            
        self.assertEqual(test_instance_1[0], stats.fisher_exact)
        self.assertEqual(test_instance_1[1], stats.mannwhitneyu)
        self.assertEqual(test_instance_2[0], self.simple_test_object.PA)
        self.assertEqual(test_instance_2[1], stats.norm.cdf)

    def test_count(self):
        '''
        test both frequency and presence absence counting in Test.
        '''
        
        self.assertEqual(self.simple_test_object.count("K00001", "group_1", False), (1,0))
        self.assertEqual(self.simple_test_object.count("K00001", "group_1", True), ([1],0))
        self.assertEqual(self.simple_test_object.count("K00001", "group_2", False), (1,1))
        self.assertEqual(self.simple_test_object.count("K00001", "group_2", True), ([0, 5], 0))
        self.assertEqual(self.simple_test_object.count("K00003", "group_2", False), (2, 0))
        self.assertEqual(self.simple_test_object.count("K00003", "group_2", True), ([1, 5], 0))

    def test_gene_frequencies(self):
        expect_1 = [['K00002', 'group_1', 'group_2', [[2], 0], [[0, 4], 0]],
                    ['K00003', 'group_1', 'group_2', [[0], 0], [[1, 5], 0]],
                    ['K00001', 'group_1', 'group_2', [[1], 0], [[0, 5], 0]]]
        expect_2 = [['K00002', 'group_1', 'group_2', [1, 0], [1, 1]],
                    ['K00003', 'group_1', 'group_2', [0, 1], [2, 0]],
                    ['K00001', 'group_1', 'group_2', [1, 0], [1, 1]]]
        for result in self.simple_test_object.gene_frequencies("group_1", "group_2", True):
            if result in expect_1:
                expect_1.pop(expect_1.index(result))
        self.assertEqual(expect_1, list())
        
        for result in self.simple_test_object.gene_frequencies("group_1", "group_2", False):
            if result in expect_2:
                expect_2.pop(expect_2.index(result))
        self.assertEqual(expect_2, list())
    
    def test_test_weighted_abundances(self):
        expect = [[[['annotation', 'group_1', 'group_2', 'enriched_in', 'group_1_mean', 'group_2_mean', 'score', 'pvalue', 'corrected_pvalue', 'description'],
                    ['K00001', 'sample_group_1', 'sample_group_2', 'sample_group_1', '23.866666666666664', '6.5', 0.0, 0.04042779918502612, '0.060591636418731595', 'E1.1.1.1, adh; alcohol dehydrogenase [EC:1.1.1.1]'],
                    ['K00002', 'sample_group_1', 'sample_group_2', 'sample_group_1', '19.733333333333334', '12.4', 0.5, 0.060591636418731595, '0.060591636418731595', 'AKR1A1, adh; alcohol dehydrogenase (NADP+) [EC:1.1.1.2]'],
                    ['K00003', 'sample_group_1', 'sample_group_2', 'sample_group_1', '24.26666666666667', '5.533333333333334', 0.0, 0.04042779918502612, '0.060591636418731595', 'hom; homoserine dehydrogenase [EC:1.1.1.3]']],
                     'sample_group_1_vs_sample_group_2_gvg_results.mannwhitneyu.tsv']]
        
        result = self.simple_test_object.test_weighted_abundances(self.sample_to_annotation,
                                                                  self.annotations)

        self.assertEqual(expect, result)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=line-too-long
"""
Various functions that apply extra annotation filters to the annotation process
"""
import logging
from itertools import product, chain
from enrichm.toolbox import window, cluster

class ClassifyChecks:
    '''
    Applies extra, manually defined filters to the 'Classify' annotation process.
    '''
    def __init__(self, classify_checks):
        self.classify_checks = classify_checks

    def check(self, module_name, genome_gff):
        """ Description
        :type self:
        :param self:

        :type module_name:
        :param module_name:

        :type genome_gff:
        :param genome_gff:

        :raises:

        :rtype:
        """
        logging.debug(f"Running rule checks for {module_name}")
        if module_name in self.classify_checks.members:
            
            if not self.check_synteny(module_name, genome_gff):
                logging.debug(f"Module {module_name} in genome failed synteny check")
                return False
            elif not self.check_required(module_name, genome_gff):
                logging.debug(f"Module {module_name} in genome failed required check")
                return False
            else:
                logging.debug(f"Module {module_name} passed all checks in genome")
                return True # in the clear

        else:
            return True

    def check_synteny(self, module_name, genome_gff):
        """ Description
        :type self:
        :param self:

        :type module_name:
        :param module_name:

        :type genome_gff:
        :param genome_gff:

        :raises:

        :rtype:
        """
        synteny_rules = self.classify_checks.get_synteny_rules(module_name)
        result_list = list()

        for parameters in synteny_rules:

            synteny_range = parameters['range']

            genes_on_contigs = dict()

            for gene in parameters['genes']:

                for contig, positions in genome_gff.items():

                    if gene in positions:

                        if contig in genes_on_contigs:
                            genes_on_contigs[contig].append(genome_gff[contig][gene])

                        else:
                            genes_on_contigs[contig] = [genome_gff[contig][gene]]

            result = False
            broke = False

            for contig, gene_positions in genes_on_contigs.items():
                
                if broke:
                    break
                for combination in product(*gene_positions):

                    endings = [gene[1] for gene in combination]
                    
                    clustered_group = list()
                    for subblock in cluster(endings, synteny_range):
                        if len(subblock) >= parameters['minsubblocksize']:
                            clustered_group.append(subblock)

                    if len(clustered_group) <= (parameters['breaks']+1):
                        if len(list(chain(*clustered_group))) >= parameters['min']:
                            result = True
                            broke = True
                            logging.debug(f"Valid syntenous block found for {module_name} ({', '.join(parameters['genes'])}) on contig {contig}")
                            break

            result_list.append(result)

        if all(result_list):
            return True
        else:
            return False

    def check_required(self, module_name, genome_gff):

        """ Description
        :type self:
        :param self:

        :type module_name:
        :param module_name:

        :type genome_gff:
        :param genome_gff:

        :raises:

        :rtype:
        """
        genes_in_genome = set(list(chain(*[list(sub.keys()) for sub in genome_gff.values()])))
        required_rules = self.classify_checks.get_required_rules(module_name)
        rule_satisfied = False

        for required_data in required_rules:

            for required_gene_set in required_data:
                number_found = 0
                for required_gene in required_gene_set:
                    if required_gene in genes_in_genome:
                        number_found += 1

                if number_found == len(required_gene_set):
                    logging.debug(f"Required gene for {module_name} ({', '.join(required_gene_set)}) found in genome")
                    rule_satisfied = True
                    break

        return rule_satisfied



#!/usr/bin/env python3
# pylint: disable=line-too-long
"""
Various functions for generating machine learning models for genomes.
"""
# Imports
import pickle
import os
import logging
import numpy as np
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.model_selection import train_test_split, RandomizedSearchCV, GridSearchCV
# Local
from enrichm.writer import Writer
from enrichm.parser import Parser

################################################################################

class GenerateModel:
    '''
    Functions to generate, train and optimise random forest machine learning models.
    '''
    def __init__(self):
        '''
        Inputs
        ------

        Outputs
        -------

        '''
        # Subparser names
        self.regressor = "regressor"
        self.classifier = "classifier"

        # Output file names
        self.attribute_importances = 'attribute_importances.tsv'
        self.model_pickle = "rf_model.pickle"
        self.labels_dict = "labels_dict.pickle"
        self.model_accuracy = "accuracy.tsv"

        # Headers
        self.attribute_importances_header = ['Variable', 'Importance']

    def numerify(self, input_list):
        '''

        Inputs
        ------

        Outputs
        -------

        '''
        counter = 0
        output_dictionary = dict()
        output_list = list()

        for group in input_list:
            group = group.pop()

            if group not in output_dictionary:
                output_dictionary[group] = counter
                counter += 1

            output_list.append(output_dictionary[group])

        output_dictionary = {item:key for key, item in output_dictionary.items()}

        return output_dictionary, np.array(output_list)

    def get_importances(self, model, attribute_list):
        '''

        Inputs
        ------

        Outputs
        -------

        '''
        importances = list(model.feature_importances_)

        # List of tuples with variable and importance
        feature_importances = [(feature, round(importance, 2)) for feature, importance in zip(attribute_list, importances)]

        # Sort the feature importances by most important first
        feature_importances = sorted(feature_importances, key=lambda x: x[1], reverse=True)

        # Print out the feature and importances
        logging.info('Writing attribute importances')

        output_lines = [self.attribute_importances_header]

        important_features = 0

        for pair in feature_importances:
            var, imp = pair
            output_lines.append([str(var), str(imp)])

            if imp > 0:
                important_features += 1

        logging.info('%i attributes found with an importance > 0', important_features)

        return output_lines

    def transpose(self, labels, features, attribute_list):
        '''

        Inputs
        ------

        Outputs
        -------

        '''
        labels_list = list()
        features_list = list()

        for col in list(labels.keys()):

            labels_list.append(labels[col])
            col_values = list()

            for attribute in attribute_list:
                col_values.append(features[col][attribute])

            features_list.append(col_values)

        return labels_list, np.array(features_list)

    def grid_search_cv(self, random_search_cv, threads, random_forest_model):
        '''
        Inputs
        ------

        Outputs
        -------

        '''
        logging.info('Generating model for grid search cross validation')

        bootstrap = [random_search_cv.best_params_['bootstrap']]
        max_depth = range(random_search_cv.best_params_['max_depth']-9, random_search_cv.best_params_['max_depth']+10, 2)
        max_features = [random_search_cv.best_params_['max_features'], 2, 3]
        min_samples_leaf = range(random_search_cv.best_params_['min_samples_leaf']-1, random_search_cv.best_params_['min_samples_leaf']+2)
        min_samples_split = range(random_search_cv.best_params_['min_samples_split']-2, random_search_cv.best_params_['min_samples_split']+3, 2)
        n_estimators = range(random_search_cv.best_params_['n_estimators']-200, random_search_cv.best_params_['n_estimators'] + 200, 100)

        param_grid = {'bootstrap':bootstrap,
                      'max_depth':[x for x in max_depth if x > 0],
                      'max_features':max_features,
                      'min_samples_leaf':[x for x in min_samples_leaf if x > 0],
                      'min_samples_split':[x for x in min_samples_split if x > 0],
                      'n_estimators':[x for x in n_estimators if x > 0]
                      }

        grid_search = GridSearchCV(estimator=random_forest_model,
                                   param_grid=param_grid,
                                   cv=3,
                                   n_jobs=threads)
        return grid_search

    def random_search_cv(self, threads, random_forest_model):
        '''
        Inputs
        ------

        Outputs
        -------

        '''
        logging.info('Generating model for random search cross validation')

        n_estimators = [int(x) for x in np.linspace(start=200, stop=2000, num=10)]
        max_features = ['auto', 'sqrt']
        max_depth = [int(x) for x in np.linspace(10, 110, num=11)]
        max_depth.append(None)
        min_samples_split = [2, 5, 10]
        min_samples_leaf = [1, 2, 4]
        bootstrap = [True, False]

        random_grid = {'n_estimators': n_estimators,
                       'max_features': max_features,
                       'max_depth': max_depth,
                       'min_samples_split': min_samples_split,
                       'min_samples_leaf': min_samples_leaf,
                       'bootstrap': bootstrap}

        rf_random = RandomizedSearchCV(estimator=random_forest_model,
                                       param_distributions=random_grid, n_iter=100, cv=3,
                                       random_state=42, n_jobs=threads)
        return rf_random

    def tune(self, features_list, labels_list_numeric, testing_portion, grid_search, threads,
             random_forest_model):
        '''
        Inputs
        ------

        Outputs
        -------

        '''
        logging.info('Splitting data into test and training datasets')
        train_features, test_features, train_labels, test_labels \
            = train_test_split(features_list,
                               labels_list_numeric,
                               test_size=testing_portion,
                               random_state=7)
        best_params_list = list()

        rf_random_model = self.random_search_cv(threads, random_forest_model)
        logging.info('Fitting model')
        rf_random_trained_model = rf_random_model.fit(train_features, train_labels)

        logging.info('Best parameters from random search cross validation:')

        for parameter, value in rf_random_trained_model.best_params_.items():
            logging.info("\t\t%s: %s", parameter, str(value))

        if grid_search:
            rf_grid_model = self.grid_search_cv(rf_random_model, threads, random_forest_model)
            logging.info('Fitting model')
            rf_grid_trained_model = rf_grid_model.fit(train_features, train_labels)

            logging.info('Best parameters from grid search cross validation:')

            for parameter, value in rf_grid_trained_model.best_params_.items():
                logging.info("\t\t%s: %s", parameter, str(value))
                best_params_list.append([parameter, str(value)])

            best_model = rf_grid_trained_model.best_estimator_

        else:
            best_model = rf_random_trained_model.best_estimator_

        return best_model, test_features, test_labels, best_params_list

    def estimate_correctness(self, predictions, test_labels):
        '''
        Calculate the percentage of predictions match with their known classifications
        (test labels).

        :param predictions: List of predictions
        :type predictions: list
        :param test_labels: List of known annotations
        :type test_labels: list
        :returns: The percentage of prediction that match their known classifications
        :rtype: float
        '''
        correctness = list()

        for prediction, label in zip(np.round(predictions), test_labels):

            if prediction == label:
                correctness.append(1)
            else:
                correctness.append(0)

        accuracy = round((sum(correctness)/float(len(correctness)))*100, 2)

        return accuracy

    def generate_pipeline(self, input_matrix_path, groups_path, model_type,
                          testing_portion, grid_search, threads, output_directory):
        '''
        Inputs
        ------

        Outputs
        -------

        '''
        logging.info('Using %f%% of the input data for testing', testing_portion*100)

        if model_type == self.regressor:
            model = RandomForestRegressor()
        elif model_type == self.classifier:
            model = RandomForestClassifier()
        else:
            raise Exception("Model type not recognised: %s" % (model_type))

        logging.info('Parsing inputs:')
        labels, _, _ = Parser.parse_metadata_matrix(groups_path)
        features, _, attribute_list = Parser.parse_simple_matrix(input_matrix_path)
        labels_list, features_list = self.transpose(labels, features, attribute_list)
        labels_dict, labels_list_numeric = self.numerify(labels_list)

        logging.info("Tuning hyperparameters")
        random_forest_model, test_features, test_labels, best_params_list = self.tune(features_list, labels_list_numeric, testing_portion, grid_search, threads, model)

        logging.info('Making predictions on test data:')
        predictions = random_forest_model.predict(test_features)
        errors = abs(predictions - test_labels)

        logging.info('\t\tMean Absolute Error: %f degrees', round(np.mean(errors), 2))
        accuracy = self.estimate_correctness(predictions, test_labels)

        logging.info('\t\tAccuracy: %f%%', accuracy)
        best_params_list.append(["Accuracy", str(accuracy)])

        logging.info("Generating attribute importances")
        output_attribute_importances = self.get_importances(random_forest_model, attribute_list)
        Writer.write(best_params_list, os.path.join(output_directory, self.model_accuracy))

        logging.info("Generating model accuracy summary file")
        Writer.write(output_attribute_importances, os.path.join(output_directory, self.attribute_importances))

        logging.info("Preserving model")
        with open(os.path.join(output_directory, self.model_pickle), 'wb') as model_io:
            pickle.dump(random_forest_model, model_io)

        logging.info("Preserving group labels")
        with open(os.path.join(output_directory, self.labels_dict), 'wb') as labels_io:
            pickle.dump(labels_dict, labels_io)



#!/usr/bin/env python3
# Imports
import os
import logging
import pickle
# Local
from enrichm.data import Data

###############################################################################

class Databases:

    def __init__(self):
        if os.path.isfile(os.path.join(Data.DATABASE_DIR, 'VERSION')):

            with open(os.path.join(Data.DATABASE_DIR, 'VERSION')) as out_io:
                self.DB_VERSION = out_io.readline().strip().replace('.tar.gz', '')

            self.CUR_DATABASE_DIR = os.path.join(Data.DATABASE_DIR, self.DB_VERSION)

            with open(os.path.join(self.CUR_DATABASE_DIR, 'VERSION')) as out_io:
                self.PICKLE_VERSION = out_io.readline().strip()

            self.IDS_DIR = os.path.join(self.CUR_DATABASE_DIR, 'ids')
            self.REF_DIR = os.path.join(self.CUR_DATABASE_DIR, 'databases')
            self.GTDB_DIR = os.path.join(self.CUR_DATABASE_DIR, 'gtdb')
            self.KO_HMM_CUTOFFS = os.path.join(self.CUR_DATABASE_DIR, 'ko_cutoffs.tsv')

            self.PICKLE = 'pickle'
            self.HMM_SUFFIX = '.hmm'
            self.DMND_SUFFIX = '.dmnd'
            self.KO_DB_NAME = 'uniref100.KO'
            self.EC_DB_NAME = 'uniref100.EC'
            self.PFAM_DB_NAME = 'pfam'
            self.KO_HMM_DB_NAME = 'ko'
            self.TIGRFAM_DB_NAME = 'tigrfam'
            self.CAZY_DB_NAME = 'cazy'
            self.GTDB_DB_NAME = 'GTDB_R80_DB'

            self.GTDB_CAZY = os.path.join(self.GTDB_DIR, "gtdb_cazy.tsv")
            self.GTDB_KO = os.path.join(self.GTDB_DIR, "gtdb_ko.tsv")
            self.GTDB_PFAM = os.path.join(self.GTDB_DIR, "gtdb_pfam.tsv")
            self.GTDB_TIGRFAM = os.path.join(self.GTDB_DIR, "gtdb_tigrfam.tsv")
            self.GTDB_EC = os.path.join(self.GTDB_DIR, "gtdb_ec.tsv")

            self.TAXONOMY = os.path.join(self.CUR_DATABASE_DIR, 'taxonomy_gtdb.tsv')
            self.M2DEF = os.path.join(self.CUR_DATABASE_DIR, 'module_to_definition')
            self.M = os.path.join(self.CUR_DATABASE_DIR, 'module_descriptions')
            self.COMPOUND_DESC = os.path.join(self.CUR_DATABASE_DIR, 'br08001')
            self.R2K = os.path.join(self.CUR_DATABASE_DIR, 'reaction_to_orthology')
            self.R2C = os.path.join(self.CUR_DATABASE_DIR, 'reaction_to_compound')
            self.R2M = os.path.join(self.CUR_DATABASE_DIR, 'reaction_to_module')
            self.M2R = os.path.join(self.CUR_DATABASE_DIR, 'module_to_reaction')
            self.M2C = os.path.join(self.CUR_DATABASE_DIR, 'module_to_cpd')
            self.R2P = os.path.join(self.CUR_DATABASE_DIR, 'reaction_to_pathway')
            self.P2R = os.path.join(self.CUR_DATABASE_DIR, 'pathway_to_reaction')
            self.C2R = os.path.join(self.CUR_DATABASE_DIR, 'compound_to_reaction')
            self.C = os.path.join(self.CUR_DATABASE_DIR, 'compound_descriptions')
            self.R = os.path.join(self.CUR_DATABASE_DIR, 'reaction_descriptions')
            self.P = os.path.join(self.CUR_DATABASE_DIR, 'pathway_descriptions')
            self.K = os.path.join(self.CUR_DATABASE_DIR, 'ko_descriptions')

            self.PFAM2CLAN = os.path.join(self.CUR_DATABASE_DIR, 'pfam_to_clan')
            self.CLAN2NAME = os.path.join(self.CUR_DATABASE_DIR, 'clan_to_name')
            self.PFAM2NAME = os.path.join(self.CUR_DATABASE_DIR, 'pfam_to_name')
            self.PFAM2DESCRIPTION = os.path.join(self.CUR_DATABASE_DIR, 'pfam_to_description')
            self.EC2DESCRIPTION = os.path.join(self.CUR_DATABASE_DIR, 'ec_to_description')
            self.TIGRFAM2DESCRIPTION = os.path.join(self.CUR_DATABASE_DIR, 'tigrfam_descriptions')
            self.CLAN2PFAM = os.path.join(self.CUR_DATABASE_DIR, 'clan_to_pfam')
        else:
            raise Exception(f"\nNo database version file found. Have you: \n\
- Installed the EnrichM database using the 'enrichm data' command?\n\
- Specified the location of the EnrichM database by exporting a \
bash variable called ENRICHM_DB? (Currently I'm looking here: {Data.DATABASE_DIR})")
        self.signature_modules = set(['M00611', 'M00612', 'M00613', 'M00614',
                                      'M00617', 'M00618', 'M00615', 'M00616',
                                      'M00363', 'M00542', 'M00574', 'M00575',
                                      'M00564', 'M00660', 'M00664', 'M00625',
                                      'M00627', 'M00745', 'M00651', 'M00652',
                                      'M00704', 'M00725', 'M00726', 'M00730',
                                      'M00744', 'M00718', 'M00639', 'M00641',
                                      'M00642', 'M00643', 'M00769', 'M00649',
                                      'M00696', 'M00697', 'M00698', 'M00700',
                                      'M00702', 'M00714', 'M00705', 'M00746'])

        self.KO_DB = os.path.join(self.REF_DIR, self.KO_DB_NAME + self.DMND_SUFFIX)
        self.EC_DB = os.path.join(self.REF_DIR, self.EC_DB_NAME + self.DMND_SUFFIX)
        self.GTDB_DB = os.path.join(self.REF_DIR, self.GTDB_DB_NAME)

        self.PFAM_DB = os.path.join(self.REF_DIR, self.PFAM_DB_NAME + self.HMM_SUFFIX)
        self.KO_HMM_DB = os.path.join(self.REF_DIR, self.KO_HMM_DB_NAME + self.HMM_SUFFIX)
        self.TIGRFAM_DB = os.path.join(self.REF_DIR, self.TIGRFAM_DB_NAME + self.HMM_SUFFIX)
        self.CAZY_DB = os.path.join(self.REF_DIR, self.CAZY_DB_NAME + self.HMM_SUFFIX)
        self.PFAM_CLAN_DB = os.path.join(self.IDS_DIR, 'PFAM_CLANS.txt')

    def m2def(self):
        logging.debug("Loading module descriptions")
        return self.load_pickle(self.M2DEF)

    def m(self):
        logging.debug("Loading reaction to pathway information")
        return self.load_pickle(self.M)

    def r2p(self):
        logging.debug("Loading pathway to reaction information")
        return self.load_pickle(self.R2P)

    def p2r(self):
        logging.debug("Loading reaction to orthology information")
        return self.load_pickle(self.P2R)

    def r2k(self):
        logging.debug("Loading reaction to module information")
        return self.load_pickle(self.R2K)

    def r2m(self):
        logging.debug("Loading module to reaction information")
        return self.load_pickle(self.R2M)

    def m2r(self):
        logging.debug("Loading module to compound information")
        return self.load_pickle(self.M2R)

    def m2c(self):
        logging.debug("Loading reaction to compound information")
        return self.load_pickle(self.M2C)

    def r2c(self):
        logging.debug("Loading compound to reaction information")
        return self.load_pickle(self.R2C)

    def c2r(self):
        logging.debug("Loading compound descriptions")
        return self.load_pickle(self.C2R)

    def c(self):
        logging.debug("Loading pathway descriptions")
        return self.load_pickle(self.C)

    def p(self):
        logging.debug("Loading reaction descriptions")
        return self.load_pickle(self.P)

    def r(self):
        logging.debug("Loading ko descriptions")
        return self.load_pickle(self.R)

    def k(self):
        logging.debug("Loading compound classifications")
        return self.load_pickle(self.K)

    def compound_desc_dict(self):
        logging.debug("Loading pfam to clan information")
        return self.load_pickle(self.COMPOUND_DESC)

    def pfam2clan(self):
        logging.debug("Loading clan descriptions")
        return self.load_pickle(self.PFAM2CLAN)

    def clan2name(self):
        logging.debug("Loading pfam names")
        return self.load_pickle(self.CLAN2NAME)

    def pfam2name(self):
        logging.debug("Loading pfam descriptions")
        return self.load_pickle(self.PFAM2NAME)

    def pfam2description(self):
        logging.debug("Loading ec descriptions")
        return self.load_pickle(self.PFAM2DESCRIPTION)

    def ec2description(self):
        logging.debug("Loading pfam hierarchy")
        return self.load_pickle(self.EC2DESCRIPTION)

    def clan2pfam(self):
        logging.debug("Loading tigrfam descriptions")
        return self.load_pickle(self.CLAN2PFAM)

    def tigrfamdescription(self):
        logging.debug("Loading reference db paths")
        return self.load_pickle(self.TIGRFAM2DESCRIPTION)

    def taxonomy(self):
        # Oh, circular dependencies.
        # Local imports suck but making another parser class would suck more
        from enrichm.parser import Parser
        return Parser.parse_taxonomy(self.TAXONOMY)

    def k2r(self):
        k2r = dict()
        for reaction, kos in self.r2k().items():
            for ko in kos:
                if ko not in k2r:
                    k2r[ko] = list()
                k2r[ko].append(reaction)
        return k2r

    def c2m(self):
        c2m = dict()

        for module, compounds in self.m2c().items():
            substrates = compounds[0]
            for substrate in substrates:
                if substrate in c2m:
                    c2m[substrate].append(module)
                else:
                    c2m[substrate] = [module]
        return c2m

    def load_pickle(self, file):

        with open('.'.join([file, self.PICKLE_VERSION, self.PICKLE]), 'rb') as file_io:
            loaded_pickle = pickle.load(file_io)

        return loaded_pickle

    def parse_ko_cutoffs(self):
        cut_ko = dict()
        out_io = open(self.KO_HMM_CUTOFFS)
        _ = out_io.readline()

        for line in out_io:
            sline = line.strip().split('\t')

            if sline[1] == '-':
                cut_ko[sline[0]] = [0.0, "NA"]
            else:
                cut_ko[sline[0]] = [float(sline[1]), sline[2]]

        return cut_ko



#!/usr/bin/env python
# Imports
import unittest
import os.path
import sys
import subprocess
import tempfile
import filecmp

path_to_script = os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','bin','enrichm')
path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data')
path_to_annotate = os.path.join(path_to_data, 'enrichm_annotate')

sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..')]+sys.path

from enrichm.enrichment import Enrichment

###############################################################################

class Tests(unittest.TestCase):

    genome_annotation_simple_example = {"genome_1": {"K00001":1,
                                                     "K00002":2},
                                        "genome_2": {"K00003":1},
                                        "genome_3": {"K00001":5,
                                                     "K00002":4,
                                                     "K00003":5}}

    genome_groups_simple_example = {"group_1": ["genome_1"],
                                "group_2": ["genome_2", "genome_3"]}

    sample_abundance = {"sample_1": {"genome_1":1.0, "genome_2":0.5, "genome_3":3.0},
                        "sample_2": {"genome_1":0.5, "genome_2":1.2, "genome_3":5.0},
                        "sample_3": {"genome_1":0.1, "genome_2":1.1, "genome_3":6.0},
                        "sample_4": {"genome_1":5.0, "genome_2":5.2, "genome_3":0.2},
                        "sample_5": {"genome_1":6.0, "genome_2":4.9, "genome_3":0.1},
                        "sample_6": {"genome_1":7.0, "genome_2":5.0, "genome_3":0.0}}

    sample_groups = {"sample_group_1": ["sample_1", "sample_2", "sample_3"],
                     "sample_group_2": ["sample_4", "sample_5", "sample_6"]}

    genomes = ["genome_1","genome_2","genome_3"]
    annotations = ["K00001", "K00002", "K00003"]
    enrichment_test_object = Enrichment()

    def test_check_annotation_type(self):
        pfam = ['PF10117']
        self.assertEqual(
            Enrichment().check_annotation_type(pfam), Enrichment.PFAM)
        cazy = ['GH42']
        self.assertEqual(
            Enrichment().check_annotation_type(cazy), Enrichment.CAZY)
        tigrfam = ['TIGR00008']
        self.assertEqual(
            Enrichment().check_annotation_type(tigrfam), Enrichment.TIGRFAM)
        ko = ['K00399']
        self.assertEqual(
            Enrichment().check_annotation_type(ko), Enrichment.KEGG)
        ec = ['1.2.3.4']
        self.assertEqual(
            Enrichment().check_annotation_type(ec), Enrichment.EC)

    def test_calculate_portions(self):
        expected = [['Annotation', 'group_1', 'group_2'],
                    ['K00001', '1.0', '0.5'],
                    ['K00002', '1.0', '0.5'],
                    ['K00003', '0.0', '1.0']]
        result = self.enrichment_test_object.calculate_portions(self.annotations,
                                                                self.genome_groups_simple_example,
                                                                self.genome_annotation_simple_example,
                                                                self.genomes,
                                                                1)
        self.assertEqual(result, expected)

    def test_enrichment_from_ko_matrix(self):

        tmp = tempfile.mkdtemp()
        expected_output = os.path.join(path_to_data, 'enrichm_enrichment_ko')
        metadata        = os.path.join(path_to_data, 'metadata.tsv')
        cmd             = '%s enrichment --annotate_output %s --metadata %s --output %s --force --ko --verbosity 1' \
                            % (path_to_script, path_to_annotate, metadata, tmp)
        subprocess.call(cmd, shell=True)

        self.assertTrue(filecmp.dircmp(tmp, expected_output))
        # The pvalues are never exact - cannot compare files directly
        #for file in os.listdir(tmp):
        #    if file         == 'enrichment.log': continue
        #    output_file     = os.path.join(tmp, file)
        #    expected_file   = os.path.join(expected_output, file)
        #    self.assertTrue(filecmp.cmp(output_file, expected_file))

    def test_enrichment_from_pfam_matrix(self):

        tmp = tempfile.mkdtemp()
        expected_output = os.path.join(path_to_data, 'enrichm_enrichment_pfam')
        metadata        = os.path.join(path_to_data, 'metadata.tsv')
        cmd             = '%s enrichment --annotate_output %s --metadata %s --output %s --force --pfam  --verbosity 1' \
                            % (path_to_script, path_to_annotate, metadata, tmp)

        subprocess.call(cmd, shell=True)

        self.assertTrue(filecmp.dircmp(tmp, expected_output))

        # The pvalues are never exact - cannot compare files directly
        #for file in os.listdir(tmp):
        #    if file         == 'enrichment.log': continue
        #    output_file     = os.path.join(tmp, file)
        #    expected_file   = os.path.join(expected_output, file)
        #    self.assertTrue(filecmp.cmp(output_file, expected_file))

    def test_weight_annotation_matrix(self):
        expected = {'sample_group_1': {'K00001': [16.0, 25.5, 30.1],
                                       'K00002': [14.0, 21.0, 24.2],
                                       'K00003': [15.5, 26.2, 31.1]},
                    'sample_group_2': {'K00001': [6.0, 6.5, 7.0],
                                       'K00002': [10.8, 12.4, 14.0],
                                       'K00003': [6.2, 5.4, 5.0]}}

        result = self.enrichment_test_object.weight_annotation_matrix(self.sample_abundance,
                                                             self.genome_annotation_simple_example,
                                                             self.sample_groups,
                                                             self.annotations)
        self.assertEqual(result, expected)

    def test_operon_enrichment(self):
        pass
        

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=line-too-long
'''
Miscellaneous functions that dont necessarily belong anywhere but are still useful
'''
import logging
import subprocess
from itertools import islice

def list_splitter(input_list, chunk_number, chunk_max):
    """
    An iterator that separates a list into a number of smaller lists
    (chunk_number). Maximum size for the sub-lists can also be
    specified (chunk_max)
    
    :param input_list: List to be split into chunks (smaller lists)
    :type input_list: list
    :param chunk_number: Number of chunks to split list into
    :type chunk_number: int
    :param chunk_max: Maximum number of entries to have in each chunk.
    :type chunk_max: int
    :returns: A list, a chunk of input_list
    :rtype: list
    """
    list_size = float(len(input_list))
    chunk_size = int(round((list_size/chunk_number), 0))

    if chunk_size > chunk_max:
        chunk_size = chunk_max
    elif chunk_size < 1:
        chunk_size = list_size

    while list_size > 0:

        if len(input_list) <= chunk_size:
            yield input_list
            del input_list
        else:
            yield input_list[:chunk_size]
            del input_list[:chunk_size]

        try:
            list_size = len(input_list)
        except NameError:
            list_size = 0

def run_command(cmd):
    '''
    Runs a command using subprocess
    :param cmd: A string of a command to run.
    :type cmd: str
    :returns: None
    :rtype: bool
    '''
    logging.debug(cmd)
    subprocess.call(cmd, shell=True)
    logging.debug('Finished')

def get_present_annotations(input_dictionary):
    '''
    Takes an input dictionary and returns keys when they have a value of > 0
    :param input_dictionary: input dictionary, with arbitrary keys and int or float values
    :type input_dictionary: dict
    :returns: list of keys with value > 0
    :rtype: list
    '''
    output_list = list()

    for key, value in input_dictionary.items():

        if value > 0:
            output_list.append(key)

    return output_list


def window(seq, n=2):
    "Stolen directly from https://stackoverflow.com/questions/6822725/rolling-or-sliding-window-iterator"
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result

def cluster(input_list, maxgap):
    '''Stolen from https://stackoverflow.com/questions/14783947/grouping-clustering-numbers-in-python'''
    input_list.sort()
    groups = [[input_list[0]]]
    for x in input_list[1:]:
        if abs(x - groups[-1][-1]) <= maxgap:
            groups[-1].append(x)
        else:
            groups.append([x])
    return groups

def reverse_dictionary_of_lists(input_dictionary):
    output_dictionary = dict()

    for key, item_list in input_dictionary.items():
        item_list = list(item_list)

        for item in item_list:

            if item not in output_dictionary:
                output_dictionary[item] = list()

            output_dictionary[item].append(key)

    return output_dictionary


#!/usr/bin/env python3
# pylint: disable=line-too-long
"""
Various functions that take genome annotations, and determine the metabolic pathways they encode.
"""
# Imports
import os
import logging
from itertools import chain
# Local
from enrichm.databases import Databases
from enrichm.module_description_parser import ModuleDescription
from enrichm.parser import Parser, RulesJson
from enrichm.writer import Writer
from enrichm.toolbox import get_present_annotations
from enrichm.classify_checks import ClassifyChecks
###############################################################################

class Classify:
    '''
    Determines which metabolic pathways are encoded by different MAGs
    '''

    def __init__(self):
        databases = Databases()
        self.signature_modules = databases.signature_modules
        self.m2def = databases.m2def()
        self.modules = databases.m()
        self.ko_output = "module_completeness.tsv"
        self.module_paths = "module_paths.tsv"
        self.aggregate_output = "aggregate_output.tsv"

    def update_with_custom_modules(self, custom_modules):
        '''
        open a file
        :param custom_modules: Path to a file containing custom modules to add to the list of 
                               default modules to use.
        :type custom_modules: str
        '''
        custom_modules_dict = dict()

        with open(custom_modules) as custom_modules_io:
            for line in custom_modules_io:
                custom_modules_dict[line.split('\t')[0]] = line.strip().split('\t')[1]

        for key in custom_modules_dict:
            self.modules[key] = 'Custom'

        return custom_modules_dict

    def classify_pipeline(self, custom_modules, cutoff, aggregate, genome_and_annotation_matrix,
                          module_rules_json, gff_file, output_directory):
        '''

        Parameters
        ----------
        custom_modules                  - string. Path to file containing custom module definitions,
                                          consistent with KEGG module nomenclature
                                          (http://www.genome.jp/kegg/module.html)
        cutoff                          - float. Fraction of a module needed in order to be included
                                          in the output.
        genome_and_annotation_matrix    - string. Path to file containing genome - annotation matrix
        output                          - string. Path to file to output results to.
        '''
        pathway = dict()
        genome_output_lines = list()

        if module_rules_json:
            cc = ClassifyChecks(RulesJson().load(module_rules_json))

        if custom_modules:
            logging.info(f'Reading in custom modules: {custom_modules}')
            modules_to_classify = self.update_with_custom_modules(custom_modules)
        else:
            modules_to_classify = self.m2def

        if gff_file:
            logging.info("Reading in annotations from an input GFF file")
            genome_to_annotation_sets = dict()
            annotation_results = dict()
            annotation_results, genome_to_annotation_sets = Parser.parse_gff(gff_file)
        else:
            logging.info("Reading in annotations from an input matrix")
            genome_to_annotation_sets, _, _ = Parser.parse_simple_matrix(genome_and_annotation_matrix)

        if aggregate:
            logging.info(f'Reading in abundances: {genome_and_annotation_matrix}')
            abundances, _, _ = Parser.parse_simple_matrix(genome_and_annotation_matrix)
            abundance_result = dict()

        logging.info(f"Read in annotations for {len(genome_to_annotation_sets)} genomes")

        output_lines = [["Genome_name", "Module_id", "Module_name", "Steps_found", "Steps_needed", "Percent_steps_found"]]

        genome_output_lines = [["Genome_name", "Module_id", "Module_name"]]

        for name, pathway_string in modules_to_classify.items():

            if name not in self.signature_modules:
                path = ModuleDescription(pathway_string)
                pathway[name] = path

                for genome, annotation_frequency in genome_to_annotation_sets.items():
                    annotations = get_present_annotations(annotation_frequency)
                    num_covered, _, _, ko_path = path.num_covered_steps(annotations)
                    num_all = path.num_steps()

                    perc_covered = num_covered / float(num_all)


                    ko_path_list = list(chain(*ko_path.values()))

                    if perc_covered >= cutoff:

                        if module_rules_json:
                            rule_check_result = cc.check(name, annotation_results[genome])
                        else:
                            rule_check_result = True

                        if rule_check_result:

                            if path.is_single_step:

                                if perc_covered != 1:

                                    if cutoff < 1:
                                        num_all = 1
                                        num_covered = 0
                                        perc_covered = 0.0

                                    else:
                                        continue

                                else:
                                    num_all = 1
                                    num_covered = 1

                            if aggregate:

                                if genome not in abundance_result:
                                    abundance_result[genome] = dict()

                                pathway_abundance = [abundances[genome][ko] for ko in ko_path_list]

                                if len(pathway_abundance)>0:
                                    pathway_average_abundance = sum(pathway_abundance) / len(pathway_abundance)
                                else:
                                    pathway_average_abundance = 0
                                abundance_result[genome][name] = pathway_average_abundance

                            genome_output_lines.append([genome, name, self.modules[name], ','.join(ko_path_list)])
                            output_line = [genome, name, self.modules[name], str(num_covered), str(num_all), str(round(perc_covered * 100, 2))]
                            output_lines.append(output_line)

        Writer.write(output_lines, os.path.join(output_directory, self.ko_output))
        Writer.write(genome_output_lines, os.path.join(output_directory, self.module_paths))

        if aggregate:
            samples = list(abundance_result.keys())
            output_lines = ['\t'.join(["ID"] + samples) + '\n']

            for module in modules_to_classify.keys():

                if module not in self.signature_modules:
                    output_line = [module]

                    for sample in samples:

                        if module in abundance_result[sample]:
                            output_line.append(str(abundance_result[sample][module]))

                        else:
                            output_line.append('0.0')
                    output_lines.append(output_line)

            Writer.write(output_lines, os.path.join(output_directory, self.aggregate_output))



#!/usr/bin/env python3
# Imports
import re
###############################################################################

KEGG = '^(K\d{5})$'
GH = '(GH\d+)'
PL = '(PL\d+)'
TIGRFAM = '^(TIGR\d{5})$'
PFAM = '(PF\d+)'
CE = '(CE\d+)'
EC = '\\d{1,2}(\\.(\\-|\\d{1,2})){3}'

class ModuleDescription:

    def __init__(self, module_description_string):
        self.module_description_string \
            = module_description_string
        self.parsed_module, self.is_single_step \
            = ModuleDescriptionParser().parse_module_string(module_description_string)

        if isinstance(self.parsed_module, ModuleDescriptionOrRelation):
            new_and = ModuleDescriptionAndRelation()
            new_and.relations = [self.parsed_module]
            self.parsed_module = new_and

    def kos(self):
        '''Return an iterable over the total list of KOs in the module'''
        r_kegg = re.compile(KEGG)
        r_gh = re.compile(GH)
        r_pl = re.compile(PL)
        r_tigrfam = re.compile(TIGRFAM)
        r_pfam = re.compile(PFAM)
        r_ce = re.compile(CE)
        r_ec = re.compile(EC)

        if any(r_kegg.findall(self.module_description_string)):
            return r_kegg.findall(self.module_description_string)
        elif any(r_gh.findall(self.module_description_string)):
            return r_gh.findall(self.module_description_string)
        elif any(r_pl.findall(self.module_description_string)):
            return r_pl.findall(self.module_description_string)
        elif any(r_tigrfam.findall(self.module_description_string)):
            return r_tigrfam.findall(self.module_description_string)
        elif any(r_pfam.findall(self.module_description_string)):
            return r_pfam.findall(self.module_description_string)
        elif any(r_ce.findall(self.module_description_string)):
            return r_ce.findall(self.module_description_string)
        elif any(r_ec.findall(self.module_description_string)):
            return r_ec.findall(self.module_description_string)

    def num_steps(self):
        if isinstance(self.parsed_module, ModuleDescriptionAndRelation):
            return len(self.parsed_module.relations)
        else:
            raise Exception("Cannot work with non-AND type modules")

    def num_covered_steps(self, ko_set):

        if isinstance(self.parsed_module, ModuleDescriptionAndRelation):
            step_cov = 0
            path_cov = 0
            reac_cov = 0
            ko_path = dict()

            for idx, m in enumerate(self.parsed_module.relations):
                step_passed, step_counts, reaction_counts, ko = m.satisfied_with(
                    ko_set, list())

                if step_passed:
                    step_cov += 1
                    path_cov += step_counts
                    ko_path[idx] = ko

                reac_cov += reaction_counts

            return step_cov, path_cov, reac_cov, ko_path

        else:
            raise Exception("Cannot work with non-AND type modules")

class GetLowestRelation:

    @staticmethod
    def get_lowest_relation(input_relation, to_fill):

        if type(input_relation) == ModuleDescriptionKoEntry:
            to_fill.append(input_relation.ko)
        else:

            for relation in input_relation.relations:
                if type(relation) == ModuleDescriptionKoEntry:
                    to_fill.append(relation.ko)
                else:
                    GetLowestRelation.get_lowest_relation(relation, to_fill)
        return to_fill

class ModuleDescriptionAndRelation:

    def satisfied_with(self, set_of_kos, kos):
        counts = 0
        step_passed = False
        reaction_counts = 0
        founds = list()

        for r in self.relations:
            found, count, reaction_count, ko = r.satisfied_with(
                set_of_kos, kos)

            if found:
                founds.append(1)
                counts += count

                for k in ko:

                    if k not in kos:
                        kos.append(k)

            reaction_counts += reaction_count

        step_passed = len(self.relations) == sum(founds)

        return step_passed, counts, reaction_counts, kos

class ModuleDescriptionPlusRelation(ModuleDescriptionAndRelation):
    pass

class ModuleDescriptionOrRelation:
    def satisfied_with(self, set_of_kos, kos):

        counts = 0
        step_passed = False
        reaction_counts = 0

        for r in self.relations:
            found, _, reaction_count, ko = r.satisfied_with(set_of_kos, kos)

            if found:
                step_passed = True
                for k in ko:
                    if k not in kos:
                        kos.append(k)
            else:
                kos_to_extract = GetLowestRelation.get_lowest_relation(
                    r, list())
                for ko in kos_to_extract:
                    if ko in kos:
                        kos.remove(ko)
            reaction_counts += reaction_count
        return step_passed, counts, reaction_counts, kos


class ModuleDescriptionKoEntry:
    def __init__(self, ko):
        self.ko = ko

    def satisfied_with(self, set_of_kos, kos):
        found = self.ko in set_of_kos
        count = (1 if found else 0)
        reaction_count = 1
        return found, count, reaction_count, [self.ko]

class ParserHelper:
    pass

class ModuleDescriptionParser:

    def correct_substrings(self, substring_list):
        fixed_substrings = list()
        for substring in substring_list:
            # Omit optional enzyme (e.g. M00372)
            # or undefined KO group (e.g. M00079)
            if substring.startswith('-'):
                continue
            # Remove redundant and definitions
            substring = substring.replace(', ', ',')
            fixed_substrings.append(substring)
        return fixed_substrings

    def parse_module_string(self, string):

        frags1 = self.split_on_space(string)
        frags1 = self.correct_substrings(frags1)
        is_single_step = len(frags1) == 1

        if is_single_step:
            if len(self.split_on_comma(frags1[0])) > 1:
                frags1 = self.split_on_comma(frags1[0])
                master_relation = ModuleDescriptionOrRelation()
            elif len(self.split_on_plus(frags1[0])) > 1:
                frags1 = self.split_on_plus(frags1[0])
                master_relation = ModuleDescriptionAndRelation()
            else:
                frags1 = self.split_on_comma(frags1[0])
                master_relation = ModuleDescriptionOrRelation()
        else:
            master_relation = ModuleDescriptionAndRelation()

        current = ParserHelper()
        current.top_relation = master_relation
        current.understuff = frags1
        stack = list([current])

        while len(stack) > 0:
            current = stack.pop()
            new_stuff = list()

            for e in current.understuff:
                if isinstance(e, str):

                    if (' ' not in e and
                        ',' not in e and
                        '+' not in e and
                        '-' not in e):
                        new_stuff.append(ModuleDescriptionKoEntry(e))

                    else:
                        frags = self.split_on_comma(e)
                        if len(frags) == 1:

                            topush = ParserHelper()
                            comma_splits = self.split_on_space(e)
                            m = None
                            
                            if len(comma_splits) == 1:
                                plus_splits = self.split_on_plus(e)
                                minus_splits = self.split_on_minus(e)

                                if len(plus_splits) > 1:
                                    m = ModuleDescriptionPlusRelation()
                                    topush.understuff = plus_splits

                                elif len(minus_splits) > 1:
                                    m = ModuleDescriptionAndRelation()
                                    topush.understuff = minus_splits[:1]

                                else:
                                    raise Exception(
                                        "Parse exception on %s" % string)

                            else:
                                m = ModuleDescriptionAndRelation()
                                topush.understuff = comma_splits

                            topush.top_relation = m
                            stack.append(topush)
                            new_stuff.append(m)

                        else:
                            m = ModuleDescriptionOrRelation()
                            topush = ParserHelper()
                            topush.top_relation = m
                            topush.understuff = frags
                            stack.append(topush)
                            new_stuff.append(m)

                else:
                    new_stuff.append(e)

            current.top_relation.relations = new_stuff

        return master_relation, is_single_step

    def split_on(self, string, characters):
        bracket_counter = 0
        fragments = list()
        current = list()
        remove_end_brackets = True

        for i in range(len(string)):
            c = string[i]

            if c == '(':
                current += c
                bracket_counter += 1

            elif c == ')':
                current += c
                bracket_counter -= 1
                if bracket_counter == 0 and i < len(string)-1:
                    remove_end_brackets = False

            elif c in characters and bracket_counter == 0:
                fragments.append(''.join(current))
                current = ''

            else:
                current += c

        fragments.append(''.join(current))

        if remove_end_brackets and string[0] == '(':

            if string[-1] != ')':
                raise Exception("Parse error")
            return self.split_on(string[1:-1], characters)

        return fragments

    def split_on_space(self, string):
        return(self.split_on(string, ' '))

    def split_on_plus(self, string):
        return(self.split_on(string, '+'))

    def split_on_comma(self, string):
        return(self.split_on(string, ','))

    def split_on_minus(self, string):
        return(self.split_on(string, '-'))



#!/usr/bin/env python
# Imports
import unittest
import os
import sys
import tempfile
import shutil
import filecmp

PATH_TO_DATA = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')] + sys.path

from enrichm.generate import GenerateModel

###############################################################################

class Tests(unittest.TestCase):

    ml_data = 'ml_data'
    sample_matrix = 'matrix.tsv'
    sample_metadata = 'metadata.tsv'
    sample_generate = 'generate_example'

    sample_matrix_path = os.path.join(PATH_TO_DATA, ml_data, sample_matrix)
    sample_metadata_path = os.path.join(PATH_TO_DATA, ml_data, sample_metadata)
    sample_generate_path = os.path.join(PATH_TO_DATA, ml_data, sample_generate)

    def test_hello_generate(self):
        tmp = tempfile.mkdtemp()
        generate_model = GenerateModel()
        generate_model.generate_pipeline(self.sample_matrix_path,
                                         self.sample_metadata_path,
                                         generate_model.classifier,
                                         0.2, # Default testing portion
                                         False, # Dont do a grid search for fine tuning
                                         2, # Threads
                                         tmp # Output directory
                                        )
        expected_files = sorted(os.listdir(tmp))
        observed_files = sorted(os.listdir(self.sample_generate_path))
        self.assertEqual(len(expected_files), len(observed_files))
        for expected_file, observed_file in zip(expected_files, observed_files):
            #expected_file_path = os.path.join(tmp, expected_file)
            #observed_file_path = os.path.join(self.sample_generate_path, observed_file)
            # Are all files present?
            self.assertEqual(expected_file, observed_file)
        shutil.rmtree(tmp)
        # Note I chose not to match files exactly here. They change because ml models are 
        # estimations and will be different every time you make them. More tests that ensure the
        # internal functions of generate will be implemented as an alternative.

if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages, Extension

with open('README.md') as readme_file:
    readme = readme_file.read()

exec(open('enrichm/version.py').read()) # loads __version__

setup(name='enrichm',
      version=__version__,
      author='Joel Boyd, Ben Woodcroft, Alexander Baker',
      author_email='joel.boyd@uqconnect.edu.au',
      description='enrichm is a toolbox for comparing the functional composition of population genomes',
      long_description_content_type='text/markdown',
      long_description=readme,
      license='GPL3+',
      keywords=["MAGs", "Population genomes", "metagenomics", "Annotation", "Comparison"],
      include_package_data=True,
      packages=find_packages(exclude='docs'),
      install_requires=('python-dateutil>=2.8.0',
                        'statsmodels>=0.8.0rc1',
                        'numpy>=1.9.1',
                        'pandas>=0.17.1',
                        'scipy>=0.17.0',
                        'biopython>=1.66',
                        'fuzzywuzzy>=0.18.0',
                        'six>=1.10.0',
                        'tempdir>=0.7.1',
                        'scikit-learn>=0.0'),
      setup_requires=['nose>=1.0'],
      test_suite='nose.collector',
      url='https://github.com/geronimp/enrichM',
      download_url='https://github.com/geronimp/enrichm/archive/%s.tar.gz' % (__version__),
      scripts=['bin/enrichm']
)





